package dev.fitko.fitconnect.client.util;

import dev.fitko.fitconnect.api.domain.model.metadata.ContentStructure;
import dev.fitko.fitconnect.api.domain.model.metadata.Hash;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.PublicServiceType;
import dev.fitko.fitconnect.api.domain.model.metadata.SignatureType;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Purpose;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.ServiceType;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.client.model.AttachmentPayload;
import dev.fitko.fitconnect.client.model.DataPayload;
import dev.fitko.fitconnect.client.model.ServiceTypePayload;
import dev.fitko.fitconnect.client.model.SubmissionPayload;
import dev.fitko.fitconnect.core.util.Strings;

import java.io.File;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.function.Function;
import java.util.stream.Collectors;

public final class SubmissionUtil {

    private SubmissionUtil(){}

    public static CreateSubmission createSubmission(final UUID destinationId, final ServiceType serviceType, final List<UUID> attachmentIdsToAnnounce) {
        return CreateSubmission.builder()
                .destinationId(destinationId)
                .announcedAttachments(attachmentIdsToAnnounce)
                .serviceType(serviceType)
                .build();
    }

    public static ServiceType buildServiceType(final ServiceTypePayload serviceTypePayload) {
        final var serviceType = new ServiceType();
        serviceType.setName(serviceTypePayload.getName());
        serviceType.setIdentifier(serviceTypePayload.getLeikaKey());
        if (!Strings.isNullOrEmpty(serviceTypePayload.getDescription())) {
            serviceType.setDescription(serviceTypePayload.getDescription());
        }
        return serviceType;
    }

    public static PublicServiceType buildPublicServiceType(final ServiceTypePayload serviceTypePayload) {
        final var publicServiceType = new PublicServiceType();
        publicServiceType.setIdentifier(serviceTypePayload.getLeikaKey());
        publicServiceType.setName(serviceTypePayload.getName());
        if (serviceTypePayload.getDescription() != null) {
            publicServiceType.setDescription(serviceTypePayload.getDescription());
        }
        return publicServiceType;
    }

    public static Data buildData(final DataPayload dataPayload) {
        final var hash = new Hash();
        hash.setContent(dataPayload.getHashedData());
        hash.setSignatureType(SignatureType.SHA_512);

        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setMimeType(dataPayload.getMimeType());
        submissionSchema.setSchemaUri(dataPayload.getSchemaUri());

        final var data = new Data();
        data.setSubmissionSchema(submissionSchema);
        data.setHash(hash);
        return data;
    }

    public static Metadata buildMetadata(final List<Attachment> attachments, final Data data, final PublicServiceType publicServiceType) {
        final var contentStructure = new ContentStructure();
        contentStructure.setAttachments(attachments);
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);
        metadata.setPublicServiceType(publicServiceType);
        return metadata;
    }

    public static CreateSubmission buildSubmissionToAnnounce(final SubmissionPayload submissionPayload) {
        final ServiceType serviceType = buildServiceType(submissionPayload.getServiceTypePayLoad());
        final List<UUID> attachmentIds = toAttachmentIds(submissionPayload.getAttachments());
        return createSubmission(submissionPayload.getDestinationId(), serviceType, attachmentIds);
    }

    public static SubmitSubmission buildSubmitSubmission(final SubmissionPayload submissionPayload, final UUID submissionId) {
        final SubmitSubmission submission = new SubmitSubmission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData(submissionPayload.getData().getEncryptedData());
        submission.setEncryptedMetadata(submissionPayload.getEncryptedMetadata());
        return submission;
    }

    public static Attachment toHashedAttachment(final AttachmentPayload attachmentPayload) {
        final var attachment = new Attachment();
        attachment.setAttachmentId(attachmentPayload.getAttachmentId());
        attachment.setPurpose(Purpose.ATTACHMENT);
        attachment.setFilename(attachmentPayload.getFile().getName());
        attachment.setMimeType(attachmentPayload.getMimeType());

        final var hash = new Hash();
        hash.setContent(attachmentPayload.getHashedData());
        hash.setSignatureType(SignatureType.SHA_512);
        attachment.setHash(hash);
        return attachment;
    }

    public static List<UUID> toAttachmentIds(final List<AttachmentPayload> attachmentPayloads) {
        return attachmentPayloads.stream()
                .map(AttachmentPayload::getAttachmentId)
                .collect(Collectors.toList());
    }

    public static List<Attachment> toAttachmentMetadata(final List<AttachmentPayload> attachmentPayloads) {
        return attachmentPayloads.stream()
                .map(SubmissionUtil::toHashedAttachment)
                .collect(Collectors.toList());
    }

    public static Function<Map.Entry<UUID, String>, AttachmentPayload> getEncryptedAttachmentPayload() {
        return attachment -> AttachmentPayload.builder()
                .encryptedData(attachment.getValue())
                .attachmentId(attachment.getKey())
                .build();
    }

    public static Function<File, AttachmentPayload> getFileAttachmentPayload() {
        return file -> AttachmentPayload.builder()
                .file(file)
                .attachmentId(UUID.randomUUID())
                .build();
    }
}
