package dev.fitko.fitconnect.client;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.jwk.KeyUse;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.jwk.gen.RSAKeyGenerator;
import dev.fitko.fitconnect.api.config.Environment;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.destination.DestinationService;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.EncryptionException;
import dev.fitko.fitconnect.api.exceptions.KeyNotRetrievedException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.exceptions.SubmissionNotCreatedException;
import dev.fitko.fitconnect.api.exceptions.ValidationException;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.client.testutil.LogCaptor;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.File;
import java.net.URI;
import java.util.*;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.*;
import static org.hamcrest.core.IsEqual.equalTo;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class SenderClientTest {

    private Sender senderMock;
    private Environment environmentMock;
    private SenderClient.Start underTest;

    private static final LogCaptor logs = new LogCaptor();

    @BeforeEach
    public void setup() {
        this.senderMock = mock(Sender.class);
        this.environmentMock = mock(Environment.class);
        this.underTest = SenderClient.build(senderMock, "", environmentMock);
        logs.clearEvents();
    }

    @Test
    void testSendSubmissionWithJsonData() throws Exception {

        // Given

        final RSAKey publicKey = generateRsaKey(4096).toPublicJWK();

        final var destinationId = UUID.randomUUID();
        final var announcedSubmission = getAnnouncedSubmission(destinationId);
        final SubmissionSchema jsonSchema = new SubmissionSchema();
        jsonSchema.setSchemaUri(URI.create("https://json.mimetype"));
        jsonSchema.setMimeType(MimeType.APPLICATION_JSON);
        final Destination destination = getDestination(destinationId, jsonSchema);

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.createSubmission(any())).thenReturn(announcedSubmission);
        when(senderMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());
        when(senderMock.getEncryptionKeyForDestination(any(), any())).thenReturn(publicKey);
        when(senderMock.validateMetadata(any(), any())).thenReturn(ValidationResult.ok());


        // When
        final var submission = underTest
                .newSubmission()
                .withAttachment(new File("src/test/resources/attachment.txt"))
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("Service", "de:fitco:test:key")
                .submit();

        // Then
        assertTrue(submission.isPresent());
        assertThat(submission.get().getSubmissionId(), equalTo(announcedSubmission.getSubmissionId()));
        assertThat(destinationId, equalTo(announcedSubmission.getDestinationId()));
    }

    @Test
    void testSendSubmissionWithXmlData() throws Exception {

        // Given

        final RSAKey publicKey = generateRsaKey(4096).toPublicJWK();

        final var destinationId = UUID.randomUUID();
        final var announcedSubmission = getAnnouncedSubmission(destinationId);
        final SubmissionSchema xmlSchema = new SubmissionSchema();
        xmlSchema.setSchemaUri(URI.create("https://xml.mimetype"));
        xmlSchema.setMimeType(MimeType.APPLICATION_XML);
        final Destination destination = getDestination(destinationId, xmlSchema);

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.createSubmission(any())).thenReturn(announcedSubmission);
        when(senderMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());
        when(senderMock.getEncryptionKeyForDestination(any(), any())).thenReturn(publicKey);
        when(senderMock.validateMetadata(any(), any())).thenReturn(ValidationResult.ok());

        // When
        final var submission = underTest.newSubmission()
                .withAttachment(new File("src/test/resources/attachment.txt"))
                .withXmlData("<xml><test>data</test></xml>")
                .withDestination(destinationId)
                .withServiceType("Service", "de:fitco:test:key")
                .submit();

        // Then
        assertTrue(submission.isPresent());
        assertThat(submission.get().getSubmissionId(), equalTo(announcedSubmission.getSubmissionId()));
        assertThat(destinationId, equalTo(announcedSubmission.getDestinationId()));
    }

    @Test
    void testMimeTypeDetection() throws Exception {

        // Given
        final UUID destinationId = setupTestMocks();
        final File file1 = new File("src/test/resources/text_attachment_with_no_filetype");
        final File file2 = new File("src/test/resources/pdf_attachment_with_no_extension");

        // When
        final var submission = underTest.newSubmission()
                .withAttachments(List.of(file1, file2))
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isPresent());
        logs.assertContains("Detected attachment mime-type text/plain");
        logs.assertContains("Detected attachment mime-type application/pdf");
    }


    @Test
    void testWithMissingData() throws Exception {

        // Given
        final UUID destinationId = setupTestMocks();

        // When
        final var submission = underTest.newSubmission()
                .withJsonData(null)
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Data is mandatory, but was null");
    }

    @Test
    void testWithMissingDestinationId() throws Exception {

        // Given
        setupTestMocks();

        // When
        final var submission = underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(null)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("DestinationId is mandatory, but was null");
    }

    @Test
    void testMissingDataMimeTypeSchema() throws Exception {

        // Given
        final var destinationId = UUID.randomUUID();
        final var destination = new Destination();
        destination.setDestinationId(destinationId);
        final var announcedSubmission = getAnnouncedSubmission(destinationId);

        final RSAKey publicKey = generateRsaKey(4096).toPublicJWK();

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.createSubmission(any())).thenReturn(announcedSubmission);
        when(senderMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());
        when(senderMock.getEncryptionKeyForDestination(any(), any())).thenReturn(publicKey);

        // When
        final var submission = underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("Service", "de:fitco:test:key")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Required schema to send valid submission not found");
    }

    @Test
    void testWithMissingServiceTypeName() throws Exception {

        // Given
        final var destinationId = setupTestMocks();

        // When
        final var submission = underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType(null, "leika:key")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Service type name is mandatory, but was null");
    }

    @Test
    void testWithMissingServiceTypeIdentifier() throws Exception {

        // Given
        final var destinationId = setupTestMocks();

        // When
        final var submission = underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("Name", null)
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Service type identifier is mandatory, but was null");
    }

    @Test
    void testWithMissingServiceTypeDescription() throws Exception {

        // Given
        final var destinationId = setupTestMocks();

        // When
        final var submission = underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("Name", null, "leika:key")
                .submit();

        // Then
        assertTrue(submission.isPresent());
    }

    @Test
    void testWithEmptyAttachments() throws Exception {

        // Given
        final UUID destinationId = setupTestMocks();

        // When
        final var submission = underTest.newSubmission()
                .withAttachments(List.of())
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isPresent());
        logs.assertContains("No attachments to upload");
    }

    @Test
    void testWithAttachmentFileNull() throws Exception {

        // Given
        final UUID destinationId = setupTestMocks();

        // When
        final var submission = underTest.newSubmission()
                .withAttachment(null)
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isPresent());
        logs.assertContains("No attachments to upload");
    }

    @Test
    void testWithAttachmentPathNotPresent() throws Exception {

        // Given
        final UUID destinationId = setupTestMocks();

        // When
        final var submission = underTest.newSubmission()
                .withAttachment(new File("/non/existing/path"))
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Reading file failed. Attachment will not be created.");
    }

    @Test
    void testInvalidPublicKeyWithSilentValidation() throws Exception {

        // Given
        final var destinationId = setupTestMocks();
        when(senderMock.validatePublicKey(any())).thenReturn(ValidationResult.error(new Exception("invalid key")));
        when(environmentMock.isSilentKeyValidation()).thenReturn(true);

        // When
        final var submission = underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isPresent());
        logs.assertContainsNot("Invalid public key");
    }

    @Test
    void testInvalidPublicKeyWithVerboseValidation() throws Exception {

        // Given
        final var destinationId = setupTestMocks();
        when(senderMock.validatePublicKey(any())).thenReturn(ValidationResult.error(new Exception("invalid key")));
        when(environmentMock.isSilentKeyValidation()).thenReturn(false);

        // When
        final var submission = underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Invalid public key");
    }

    @Test
    void testEncryptionError() throws Exception {

        // Given
        final var destinationId = setupTestMocks();
        when(senderMock.encryptBytes(any(),any())).thenThrow(new EncryptionException("Encryption failed"));

        // When
        final var submission = underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Encrypting submission failed");
    }

    @Test
    void testAnnouncingSubmissionFailed() throws Exception {

        // Given
        final var destinationId = setupTestMocks();
        when(senderMock.createSubmission(any())).thenThrow(new SubmissionNotCreatedException("Announcing submission failed"));

        // When
        final var submission = underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Failed to announce new submission");

    }

    @Test
    void testRestApiCallFailed() throws Exception {

        // Given
        final var destinationId = setupTestMocks();
        when(senderMock.getDestination(any())).thenThrow(new RestApiException("Loading destination failed"));

        // When
        final Optional<SubmitSubmission> submission = underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        assertTrue(submission.isEmpty());
        logs.assertContains("Sending submission failed");

    }

    @Test
    void testGettingEncryptionKeyFailed() throws Exception {

        // Given
        final var destinationId = setupTestMocks();
        when(senderMock.getEncryptionKeyForDestination(any(), any())).thenThrow(new KeyNotRetrievedException("Getting encryption key failed"));

        // When
        final Optional<SubmitSubmission> submission = underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        assertTrue(submission.isEmpty());
        logs.assertContains("Getting encryption key for destination "+ destinationId+" failed");
    }

    @Test
    void testSendSubmissionWithEncryptedData() throws Exception {

        // Given
        final var destinationId = setupTestMocks();

        // When
        final Optional<SubmitSubmission> submission = underTest.newSubmissionWithEncryptedData()
                .withEncryptedAttachments(Map.of(UUID.randomUUID(),"@tt@chm$nt"))
                .withEncryptedJsonData("$ncrypt$d d@t@")
                .withEncryptedMetadata("$ncrypt$d m$t@d@t@")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        assertTrue(submission.isPresent());
        logs.assertContains("SUCCESSFULLY HANDED IN SUBMISSION");
    }

    @Test
    void testPublicEncryptionKeyRetrieval() throws Exception {

        // Given
        final var destinationId = UUID.randomUUID();
        final var destination = getDestination(destinationId);

        final RSAKey publicKey = generateRsaKey(4096).toPublicJWK();

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());
        when(senderMock.getEncryptionKeyForDestination(any(), any())).thenReturn(publicKey);

        // When
        final Optional<String> encryptionKey = underTest.getPublicKey(destinationId);

        assertTrue(encryptionKey.isPresent());
        assertThat(encryptionKey.get(), equalTo(publicKey.toJSONString()));
    }

    @Test
    void testFailOnInvalidMetadata() throws Exception {

        // Given
        final var destinationId = setupTestMocks();
        when(senderMock.validateMetadata(any(), any())).thenReturn(ValidationResult.error(new ValidationException("invalid metadata")));

        // When
        final Optional<SubmitSubmission> submission = underTest.newSubmission()
                .withJsonData("{}")
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .submit();

        assertTrue(submission.isEmpty());
        logs.assertContains("Metadata does not match schema");
    }

    private UUID setupTestMocks() throws JOSEException {
        final var destinationId = UUID.randomUUID();
        final var destination = getDestination(destinationId);
        final var announcedSubmission = getAnnouncedSubmission(destinationId);

        final RSAKey publicKey = generateRsaKey(4096).toPublicJWK();

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.createSubmission(any())).thenReturn(announcedSubmission);
        when(senderMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());
        when(senderMock.getEncryptionKeyForDestination(any(), any())).thenReturn(publicKey);
        when(senderMock.validateMetadata(any(), any())).thenReturn(ValidationResult.ok());

        return destinationId;
    }

    private SubmissionForPickup getAnnouncedSubmission(final UUID destinationId) {
        final SubmissionSchema submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://json-schema.fitco.de"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final DestinationService destinationService = new DestinationService();
        destinationService.setIdentifier("1");
        destinationService.setSubmissionSchemas(Set.of(submissionSchema));

        final var announcedSubmission = new SubmissionForPickup();
        announcedSubmission.setSubmissionId(UUID.randomUUID());
        announcedSubmission.setDestinationId(destinationId);
        return announcedSubmission;
    }

    private Destination getDestination(final UUID destinationId) {
        final SubmissionSchema schema = new SubmissionSchema();
        schema.setSchemaUri(URI.create("https://mimetype.test.de"));
        schema.setMimeType(MimeType.APPLICATION_JSON);

        final var destinationService = new DestinationService();
        destinationService.setSubmissionSchemas(Set.of(schema));

        final var destination = new Destination();
        destination.setDestinationId(destinationId);
        destination.setServices(Set.of(destinationService));
        return destination;
    }

    private Destination getDestination(final UUID destinationId, final SubmissionSchema schema) {
        final var destinationService = new DestinationService();
        destinationService.setSubmissionSchemas(Set.of(schema));

        final var destination = new Destination();
        destination.setDestinationId(destinationId);
        destination.setServices(Set.of(destinationService));
        return destination;
    }

    public static RSAKey generateRsaKey(final int size) throws JOSEException {
        return new RSAKeyGenerator(size)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate();
    }
}