package dev.fitko.fitconnect.client;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.domain.model.metadata.ContentStructure;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.metadata.SignatureType;
import dev.fitko.fitconnect.api.domain.model.metadata.Hash;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.DecryptionException;
import dev.fitko.fitconnect.api.exceptions.EventCreationException;
import dev.fitko.fitconnect.api.exceptions.InvalidSigningKeyException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.exceptions.ValidationException;
import dev.fitko.fitconnect.api.services.Subscriber;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.client.testutil.LogCaptor;
import dev.fitko.fitconnect.core.crypto.HashService;
import dev.fitko.fitconnect.core.crypto.JWECryptoService;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.nio.file.Files;
import java.text.ParseException;
import java.util.*;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;

class SubscriberClientTest {

    public static final ObjectMapper mapper = new ObjectMapper();

    private final LogCaptor logs = new LogCaptor();

    private Subscriber subscriberMock;
    private String privateKey;
    private SubscriberClient.RequestSubmission underTest;

    @BeforeEach
    public void setup() throws IOException {
        privateKey = getResourceAsString("private_decryption_test_key.json");
        final String schema = getResourceAsString("metadata_schema.json");
        subscriberMock = Mockito.mock(Subscriber.class);
        underTest = SubscriberClient.builder(subscriberMock, privateKey, schema);
    }

    @Test
    void testMissingPrivateKey() {
        assertNull(SubscriberClient.builder(subscriberMock, null, ""));
        logs.assertContains("No decryption key set, please check environment");
    }

    @Test
    void testPrivateKeyNotParseable() {
        // Given
        final var client = SubscriberClient.builder(subscriberMock, "broken key", "");

        // When
        final var receivedSubmission = client.requestSubmission(UUID.randomUUID());

        // Then
        assertNotNull(client);
        assertTrue(receivedSubmission.isEmpty());
        logs.assertContains("Private key could not be parsed");
    }

    @Test
    void testMissingMetadataSchema() {
        assertNull(SubscriberClient.builder(subscriberMock, privateKey, null));
        logs.assertContains("No metadata schema present, please check environment");
    }

    @Test
    void testNoSubmissionsFound() {
        assertThat(underTest.getAvailableSubmissions(null).isEmpty(), is(true));
        assertThat(underTest.getAvailableSubmissions(UUID.randomUUID()).isEmpty(), is(true));

        final var destinationId = UUID.randomUUID();
        when(subscriberMock.pollAvailableSubmissions(destinationId, 0, 5)).thenReturn(Collections.EMPTY_SET);
        assertThat(underTest.getAvailableSubmissions(destinationId).isEmpty(), is(true));
    }

    @Test
    void testRestCallFailed() {
        // Given
        when(subscriberMock.getSubmission(any())).thenThrow(new RestApiException("Submission not found"));

        // When
        final var receivedSubmission = underTest.requestSubmission(UUID.randomUUID());

        // Then
        assertTrue(receivedSubmission.isEmpty());
        logs.assertContains("API request failed");
    }

    @Test
    void testDecryptionFailed() {
        // Given
        final var submissionId = UUID.randomUUID();
        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData("encryptedData");
        submission.setEncryptedMetadata("encryptedMetadata");

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.decryptStringContent(any(), any())).thenThrow(new DecryptionException("Decryption failed"));

        // When
        final var receivedSubmission = underTest.requestSubmission(submissionId);

        // Then
        assertTrue(receivedSubmission.isEmpty());
        logs.assertContains("Decrypting payload failed");
    }

    @Test
    void testCreatingSetEventFailed() throws JsonProcessingException {
        // Given
        final var submissionId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);

        final var metadata = new Metadata();

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.decryptStringContent(any(), any())).thenReturn(mapper.writeValueAsBytes(metadata));
        when(subscriberMock.validateMetadata(any(), any())).thenReturn(ValidationResult.error(new Exception()));

        doAnswer(invocation -> {
            throw new EventCreationException("");
        }).when(subscriberMock).rejectSubmission(any(), any(), any());

        // When
        final var receivedSubmission = underTest.requestSubmission(submissionId);

        // Then
        assertTrue(receivedSubmission.isEmpty());
        logs.assertContains("Creating SET event failed");
    }

    @Test
    void testSigningKeyInvalid() throws JsonProcessingException {
        // Given
        final var submissionId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);

        final var metadata = new Metadata();

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.decryptStringContent(any(), any())).thenReturn(mapper.writeValueAsBytes(metadata));
        when(subscriberMock.validateMetadata(any(), any())).thenReturn(ValidationResult.error(new Exception()));

        doAnswer(invocation -> {
            throw new InvalidSigningKeyException("");
        }).when(subscriberMock).rejectSubmission(any(), any(), any());

        // When
        final var receivedSubmission = underTest.requestSubmission(submissionId);

        // Then
        assertTrue(receivedSubmission.isEmpty());
        logs.assertContains("Signing the SET event failed");
    }



    @Test
    void testReadingMetadataFailed() {
        // Given
        final var submissionId = UUID.randomUUID();
        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData("encryptedData");
        submission.setEncryptedMetadata("encryptedMetadata");

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.decryptStringContent(any(), any())).thenReturn("encryptedMetadata".getBytes());

        // When
        final var receivedSubmission = underTest.requestSubmission(submissionId);

        // Then
        assertTrue(receivedSubmission.isEmpty());
        logs.assertContains("Reading metadata failed");
    }


    @Test
    void testSubmissionsFound() {
        // Given
        final var destinationId = UUID.randomUUID();

        final SubmissionForPickup submission = new SubmissionForPickup();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(destinationId);

        when(subscriberMock.pollAvailableSubmissions(destinationId, 0, 1)).thenReturn(Set.of(submission));

        // When
        final Set<SubmissionForPickup> availableSubmissions = underTest.getAvailableSubmissions(destinationId, 0, 1);

        // Then
        assertThat(availableSubmissions.size(), is(1));
        assertThat(availableSubmissions.stream().findFirst().get(), is(submission));
    }

    @Test
    void testGetSubmission() throws Exception {

        // Given
        final var dataPayload = "{ some : 'data foo' }";
        final var submissionId = UUID.randomUUID();
        final CryptoService cryptoService = new JWECryptoService(new HashService());

        final Hash hash = new Hash();
        hash.setContent(cryptoService.hashBytes(dataPayload.getBytes()));
        hash.setSignatureType(SignatureType.SHA_512);

        final SubmissionSchema schema = new SubmissionSchema();
        schema.setSchemaUri(URI.create("https://dummy.schema.url"));
        schema.setMimeType(MimeType.APPLICATION_JSON);

        final Data data = new Data();
        data.setSubmissionSchema(schema);
        data.setHash(hash);

        final ContentStructure contentStructure = new ContentStructure();
        contentStructure.setAttachments(List.of());
        contentStructure.setData(data);

        final Metadata metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        final byte[] metadataBytes = mapper.writeValueAsBytes(metadata);

        final RSAKey encryptionKey = RSAKey.parse(privateKey);
        final RSAKey decryptionKey = RSAKey.parse(privateKey).toPublicJWK();

        final String encryptedData = cryptoService.encryptString(decryptionKey, dataPayload);
        final String encryptedMetadata = cryptoService.encryptBytes(decryptionKey, metadataBytes);

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData(encryptedData);
        submission.setEncryptedMetadata(encryptedMetadata);

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.validateMetadata(any(), any())).thenReturn(ValidationResult.ok());
        when(subscriberMock.validateHashIntegrity(any(), any())).thenReturn(ValidationResult.ok());
        when(subscriberMock.decryptStringContent(encryptionKey, submission.getEncryptedData())).thenReturn(dataPayload.getBytes());
        when(subscriberMock.decryptStringContent(encryptionKey, submission.getEncryptedMetadata())).thenReturn(metadataBytes);

        // When
        final var receivedSubmission = underTest.requestSubmission(submissionId);

        // Then
        assertTrue(receivedSubmission.isPresent());
        assertThat(receivedSubmission.get().getData(), is(dataPayload));
        assertThat(receivedSubmission.get().getMimeType(), is(MimeType.APPLICATION_JSON));
        assertThat(receivedSubmission.get().getAttachments(), is(Collections.EMPTY_LIST));
    }

    @Test
    void testSubmissionNotFound() {
        // Given
        when(subscriberMock.getSubmission(any())).thenReturn(null);

        // When
        final var receivedSubmission = underTest.requestSubmission(UUID.randomUUID());

        // Then
        assertTrue(receivedSubmission.isEmpty());
    }

    @Test
    void testInvalidMetadata() throws JsonProcessingException {
        // Given
        final var submissionId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setDestinationId(destinationId);
        submission.setCaseId(caseId);

        final var invalidMetadata = mapper.writeValueAsBytes(new Metadata());

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.decryptStringContent(any(), any())).thenReturn(invalidMetadata);
        when(subscriberMock.validateMetadata(any(), any())).thenReturn(ValidationResult.error(new ValidationException("Validation failed")));

        // When
        final var receivedSubmission = underTest.requestSubmission(submissionId);

        // Then
        assertTrue(receivedSubmission.isEmpty());
        verify(subscriberMock).rejectSubmission(submissionId, destinationId, caseId);
        logs.assertContains("Metadata does not match schema");
    }

    @Test
    void testCorruptedData() throws JsonProcessingException, ParseException {
        // Given
        final RSAKey decryptionKey = RSAKey.parse(privateKey);
        final RSAKey encryptionKey = decryptionKey.toPublicJWK();
        final CryptoService cryptoService = new JWECryptoService(new HashService());

        final var dataPayload = "some data that was transferred";

        final var hash = new Hash();
        hash.setContent(cryptoService.hashBytes(dataPayload.getBytes()));
        final var data = new Data();
        data.setHash(hash);
        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);
        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        final String encryptedData = cryptoService.encryptString(encryptionKey, mapper.writeValueAsString(data));
        final String encryptedMetadata = cryptoService.encryptString(encryptionKey, mapper.writeValueAsString(metadata));
        final var metadataBytes = mapper.writeValueAsBytes(metadata);
        final var dataBytes = mapper.writeValueAsBytes(data);

        final var submissionId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setCaseId(caseId);
        submission.setSubmissionId(submissionId);
        submission.setDestinationId(destinationId);
        submission.setEncryptedMetadata(encryptedMetadata);
        submission.setEncryptedData(encryptedData);

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.decryptStringContent(decryptionKey, encryptedMetadata)).thenReturn(metadataBytes);
        when(subscriberMock.decryptStringContent(decryptionKey, encryptedData)).thenReturn(dataBytes);
        when(subscriberMock.validateMetadata(any(), any())).thenReturn(ValidationResult.ok());
        when(subscriberMock.validateHashIntegrity(any(), any())).thenReturn(ValidationResult.error(new ValidationException("Corrupt data hash")));

        // When
        final var receivedSubmission = underTest.requestSubmission(submissionId);

        // Then
        assertTrue(receivedSubmission.isEmpty());
        verify(subscriberMock).rejectSubmission(submissionId, destinationId, caseId);
        logs.assertContains("Data might be corrupted, hash-sum doe not match");
    }

    @Test
    void testCorruptedAttachment() throws JsonProcessingException, ParseException {
        // Given
        final RSAKey decryptionKey = RSAKey.parse(privateKey);
        final RSAKey encryptionKey = decryptionKey.toPublicJWK();
        final CryptoService cryptoService = new JWECryptoService(new HashService());

        final var dataHash = new Hash();
        dataHash.setContent("");
        final var data = new Data();
        data.setHash(dataHash);

        final var attachmentHash = new Hash();
        attachmentHash.setContent("");
        final var attachment = new Attachment();
        attachment.setFilename("src/test/resources/attachment.txt");
        attachment.setHash(attachmentHash);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);
        contentStructure.setAttachments(List.of(attachment));
        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        final String encryptedData = cryptoService.encryptString(encryptionKey, mapper.writeValueAsString(data));
        final String encryptedMetadata = cryptoService.encryptString(encryptionKey, mapper.writeValueAsString(metadata));
        final var metadataBytes = mapper.writeValueAsBytes(metadata);
        final var dataBytes = mapper.writeValueAsBytes(data);

        final var submissionId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setCaseId(caseId);
        submission.setSubmissionId(submissionId);
        submission.setDestinationId(destinationId);
        submission.setEncryptedMetadata(encryptedMetadata);
        submission.setEncryptedData(encryptedData);

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.decryptStringContent(decryptionKey, encryptedMetadata)).thenReturn(metadataBytes);
        when(subscriberMock.decryptStringContent(decryptionKey, encryptedData)).thenReturn(dataBytes);
        when(subscriberMock.validateMetadata(any(), any())).thenReturn(ValidationResult.ok());

        final var decryptedDataBytes = cryptoService.decryptBytes(decryptionKey, submission.getEncryptedData());
        when(subscriberMock.validateHashIntegrity(dataHash.getContent(), decryptedDataBytes)).thenReturn(ValidationResult.ok());
        when(subscriberMock.validateHashIntegrity(attachmentHash.getContent(), null)).thenReturn(ValidationResult.error(new ValidationException("Corrupt attachment")));

        // When
        final var receivedSubmission = underTest.requestSubmission(submissionId);

        // Then
        assertTrue(receivedSubmission.isEmpty());
        verify(subscriberMock).rejectSubmission(submissionId, destinationId, caseId);
        logs.assertContains("Attachment data for id null is corrupted");
    }

    private String getResourceAsString(final String name) throws IOException {
        final ClassLoader classLoader = getClass().getClassLoader();
        final File file = new File(classLoader.getResource(name).getFile());
        return Files.readString(file.toPath());
    }
}