package dev.fitko.fitconnect.client.cmd;

import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.AttachmentWithData;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.client.SenderClient;
import dev.fitko.fitconnect.client.SubscriberClient;
import dev.fitko.fitconnect.client.model.ReceivedSubmission;
import dev.fitko.fitconnect.client.testutil.LogCaptor;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.io.TempDir;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.nio.file.Path;
import java.util.*;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class CommandLineClientTest {

    private final LogCaptor logs = new LogCaptor();

    private final ByteArrayOutputStream sysoutCaptor = new ByteArrayOutputStream();

    private CommandLineClient underTest;
    private SenderClient.ClientBuilder senderClientMock;
    private SubscriberClient.ClientBuilder subscriberClientMock;

    @BeforeEach
    void setUp() {
        senderClientMock = mock(SenderClient.ClientBuilder.class);
        subscriberClientMock = mock(SubscriberClient.ClientBuilder.class);
        underTest = new CommandLineClient(senderClientMock, subscriberClientMock);
    }

    @Test
    void testInvalidDestinationId() throws IOException {
        // When
        underTest.run("list", "--destinationId=123");

        // Then
        logs.assertContains("Parameter 123 is not a valid UUID.");
    }

    @Test
    void testListAllSubmissions() throws IOException {

        // Given

        // capture sysouts from the cmd line output
        System.setOut(new PrintStream(sysoutCaptor));

        final var destinationId = UUID.randomUUID();

        final Set<SubmissionForPickup> submissions = generateSubmissions(destinationId, 10);
        when(subscriberClientMock.getAvailableSubmissions(destinationId)).thenReturn(submissions);

        // When
        underTest.run("list", "--destinationId=" + destinationId);

        // Then
        logs.assertContains("Listing available submissions for destination " + destinationId);
        submissions.forEach(submission -> assertThat(sysoutCaptor.toString(), containsString(submission.getSubmissionId().toString())));
    }

    @Test
    void testGetSingleSubmission(@TempDir final Path tempDir) throws IOException {

        // Given
        final var submissionId = UUID.randomUUID();

        final var attachmentMetadata = new Attachment();
        attachmentMetadata.setFilename("testAttachment.txt");
        final AttachmentWithData attachment = AttachmentWithData.builder()
                .decryptedData("attachment text".getBytes())
                .attachmentMetadata(attachmentMetadata)
                .build();

        final var receivedSubmission = ReceivedSubmission.builder()
                .data("{ data: '' }")
                .attachments(List.of(attachment))
                .mimeType(MimeType.APPLICATION_JSON)
                .build();
        when(subscriberClientMock.requestSubmission(submissionId)).thenReturn(Optional.of(receivedSubmission));

        // When
        underTest.run("get", "--submissionId=" + submissionId, "--target=" + tempDir);

        // Then
        logs.assertContains("Listing submission for id " + submissionId);
        logs.assertContains("Creating data directory for submission in " + tempDir);
        logs.assertContains("Writing data.json");
        logs.assertContains("Writing attachment testAttachment.txt");
    }

    @Test
    void testSendSubmission() throws IOException {

        // Given
        final var destinationId = UUID.randomUUID();

        final var expectedSubmission = new SubmitSubmission();
        expectedSubmission.setSubmissionId(UUID.randomUUID());
        expectedSubmission.setEncryptedMetadata("encryptedMetadata");
        expectedSubmission.setEncryptedData("encryptedData");

        when(senderClientMock.withDestination(any())).thenReturn(senderClientMock);
        when(senderClientMock.withServiceType(any(), any())).thenReturn(senderClientMock);
        when(senderClientMock.withAttachments(any())).thenReturn(senderClientMock);
        when(senderClientMock.withJsonData(any())).thenReturn(senderClientMock);
        when(senderClientMock.submit()).thenReturn(Optional.of(expectedSubmission));
        when(senderClientMock.newSubmission()).thenReturn(senderClientMock);

        // When
        underTest.run("send",
                "--destinationId=" + destinationId,
                "--serviceName=Test",
                "--leikaKey=urn:de:fim:leika:leistung:99400048079000",
                "--attachments=src/test/resources/attachment.txt",
                "--data=src/test/resources/data.json",
                "--dataType=JSON");

        // Then
        logs.assertContains("Sending new submission to destination " + destinationId);
    }

    private Set<SubmissionForPickup> generateSubmissions(final UUID destinationId, final int count) {
        final Set<SubmissionForPickup> submissions = new HashSet<>(count);
        for (int i = 0; i < count; i++) {
            final SubmissionForPickup submission = new SubmissionForPickup();
            submission.setSubmissionId(UUID.randomUUID());
            submission.setDestinationId(destinationId);
            submissions.add(submission);
        }
        return submissions;
    }
}